/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This software is Copyright  2011 Yuri Gonzaga Gonalves da Costa <yuriggc at gmail>, and it is hereby released to the general public under the following terms:*
 * Redistribution and use in source and binary forms, with or without modification, are permitted.                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
`timescale 1 ns/ 1 ns
module eksblowfish_loop_2(reset, clk, VALID_E, READY_E, VALID_S, READY_S, data_in, code, data_out_a,data_out_b,index_out);

	parameter DATA_WIDTH=32;
	parameter ADDR_WIDTH=10;

	parameter IN_HANDSHAKE			= 5'b00001;
	parameter FEISTEL				= 5'b00010;
	parameter F1					= 5'b00011;
	parameter PREF 					= 5'b00100;
	parameter XOR_P17_P18			= 5'b00101;
	parameter OUT_HANDSHAKE			= 5'b00110;
	parameter NOP1					= 5'b00111;
	parameter XOR_XR_SWAP			= 5'b01000;
	parameter SWAPR					= 5'b01001;
	parameter NOP2					= 5'b01010;
	parameter INIT_XOR				= 5'b01011;
	parameter CONTROL_EXPAND_KEY	= 5'b01100;
	parameter INIT_PARRAY			= 5'b01101;
	parameter INIT_SBOXES			= 5'b01110;
	parameter CYCLIC_KEY			= 5'b01111;
	parameter F2					= 5'b10000;
	parameter F0					= 5'b10001;
	parameter PRE_OUT_HANDSHAKE   	= 5'b10010;
	parameter F0_1					= 5'b10011;
	parameter F2_1					= 5'b10100;
	
	
	
	parameter XOR_PARRAY_1	    = 4'b0000;
	parameter XOR_PARRAY_2	    = 4'b0001;
	parameter XOR_SBOX_1		= 4'b0010;
	parameter XOR_SBOX_2		= 4'b0011;
	parameter COST_1			= 4'b0100;
	parameter COST_2			= 4'b0101;
	parameter COST_3			= 4'b0110;
	parameter POS_XOR_SBOX_2	= 5'b0111;
	
	

	input reset;
	input clk;

	/*Input Handshake*/	
	input VALID_E;
	output READY_E;
	reg READY_E;
    
	/*Output Handshake*/
	output VALID_S;
	reg VALID_S;
	input  READY_S; 
	
	
	input [31:0]data_in;/*Todos os dados reais de entrada so recebidos por data_in*/
	input [2:0]code;/*Especifica a entrada*/
	
	/**
	 * data_out_a e data_out_b sao as saidas referentes as sboxes
	 */
	output [31:0]data_out_a;
	reg [31:0] data_out_a;
	output [31:0]data_out_b;
	reg [31:0] data_out_b;
	output [9:0]index_out;
	reg [9:0] index_out;
	reg [9:0] index;
	
	
	/*Registradores internos*/
	reg [31:0]L;
	reg [31:0]R;
	reg [31:0]F_r;
	
	reg [31:0]key_r[17:0];
	reg [31:0]parray[17:0];
	
	reg [6:0]n_key_r;
	reg [6:0]n_key_r2;
    reg [6:0]n_ciclic_key;
	reg [31:0]salt_r[3:0];
	reg [31:0]cost_r;
	reg ctrl; 
	
	reg [63:0]pApB;	
		
	
	reg [3:0] count_round;
	reg [4:0] count_XOR_PARRAY;
	reg [9:0] count_XOR_SBOX;
	
	reg S1S2;
	
	
	/* State machines states*/
	reg [4:0] state = IN_HANDSHAKE;
	reg [3:0] controlstate;

	/*RAM*/
	reg [(DATA_WIDTH-1):0] data_a,data_b;
	reg [(ADDR_WIDTH-1):0] addr_a,addr_b;
	reg we_a,we_b;
	
	wire clk_a,clk_b;
	wire [(DATA_WIDTH-1):0] q_a,q_b;
	
	RAM ramb(data_a,data_b,addr_a,addr_b,we_a, we_b, clk_a, clk_b,q_a, q_b);
	
	/*ROM*/
	reg [(ADDR_WIDTH-1):0] addr_ar; 
	wire [(DATA_WIDTH-1):0] q_ar;
	
	assign clk_a = clk;
	assign clk_b = clk;
	
	reg accessingMEM;
	
	integer i;
	
	always@(posedge clk) begin
		if(reset==1'b1) begin
			if(accessingMEM == 1'b0)begin
				if(state == IN_HANDSHAKE) begin//ready to read input data
					if(VALID_E ==1'b1)begin//valid data to be read
						if(code == 3'd0)begin
							cost_r <= 32'd1<<data_in;//cost
							
							ctrl <= 1'b0;
							S1S2 <= 1'b0;
							
							/*Informaoes para ler os dois primeiros dados da rom 01*/
							addr_ar <= 10'd0;
							
							we_a <= 1'b0;
							/*Informaoes para ler os dois primeiros dados da rom 01*/
							
							count_round <= 4'd0;
							
						end 
						else if(code == 3'd1)begin
							salt_r[0] <= data_in;
						end
						else if(code == 3'd2)begin
							salt_r[1]   <= data_in;
						end
						else if(code == 3'd3)begin
							salt_r[2] <= data_in;
						end
						else if(code == 3'd4)begin
							salt_r[3]   <= data_in;						
							n_key_r  <= 7'd0 ;
							n_key_r2 <= 7'd0;
						end
						else if(code == 3'd5)begin/*Codigo para passagem da chave ja extendida pelo software*/
							key_r[n_key_r]  <= data_in;
							n_key_r <= n_key_r + 7'd1 ;
						end 
						else if(code == 3'd6)begin
							parray[n_key_r2] <= data_in;
							n_key_r2 <= n_key_r2 + 7'd1;
						end 
						else if(code == 3'd7)begin
							/*Endereco a ser escrito na ram*/
							addr_a <= addr_ar;
							/*dados a serem escritosas ra,*/
							data_a <= data_in;
							/*Autoriza escrita na ram*/
							we_a   <= 1'b1;
							/*No mesmo ciclo eh feita a leitura da rom e escrita na ram*/
							READY_E <= 1'b0;
							accessingMEM <= 1'b1;
							if( addr_ar == 10'd1023) begin/*Ultimo valor para a RAM*/
								addr_ar <= 10'd0;
								/*Apos ultima escrita na ram, ela estara inicializada*/
								state <= CONTROL_EXPAND_KEY; 
							    controlstate <= COST_1;
							end
							else begin											
								/*Novos valores a serem lidos da ROM*/
								addr_ar <= addr_ar + 10'd1;
							end
						end
					end
				end
				else if(state == INIT_XOR) begin/*Xor de parray com a chave extendida*/
					if(count_XOR_PARRAY < 5'd18)begin
						if(ctrl==1'b1)begin
							parray[count_XOR_PARRAY] <= parray[count_XOR_PARRAY] ^ key_r[count_XOR_PARRAY];
						end
						else begin
							parray[count_XOR_PARRAY] <= parray[count_XOR_PARRAY] ^ salt_r[count_XOR_PARRAY%5'd4];
						end
						count_XOR_PARRAY <= count_XOR_PARRAY + 5'd1;
					end
					else begin
						$display("ctrl %b\n",ctrl);
						count_XOR_PARRAY <= 5'd0;
						pApB <= 64'h0000000000000000;
						controlstate <= XOR_PARRAY_1;
						state <= CONTROL_EXPAND_KEY;
					end
				end
				else if(state == CONTROL_EXPAND_KEY) begin
					if(controlstate == XOR_PARRAY_1) begin

						L <= pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R <= pApB[31:0];
						
						
						count_round <= 4'd0;
						controlstate <= XOR_PARRAY_2;
						state <= PREF;
					end
					else if(controlstate == XOR_PARRAY_2)begin
						parray[count_XOR_PARRAY] <= pApB[63:32];
						parray[count_XOR_PARRAY+5'd1] <= pApB[31:0];
						if(count_XOR_PARRAY < 5'd16) begin
							count_XOR_PARRAY <= count_XOR_PARRAY+5'd2;
							controlstate <= XOR_PARRAY_1;
						end
						else begin
							count_XOR_SBOX <= 10'd0;
							controlstate <= XOR_SBOX_1;
						end
					end
					else if( controlstate == XOR_SBOX_1)begin
						we_a   <= 1'b0;
						we_b   <= 1'b0;

						L 		<= pApB[63:32];//xL = xL ^ pArray[count_pArray];
						R 		<= pApB[31:0];
							
						count_round <= 4'd0;
						controlstate <= XOR_SBOX_2;
						state <= PREF;
					end
					else if(controlstate == XOR_SBOX_2)begin
						/*Atualizacao de sboxes*/
						addr_a <= count_XOR_SBOX;
						addr_b <= count_XOR_SBOX+10'd1;
						data_a <= pApB[63:32];
						data_b <= pApB[31:0];
						we_a   <= 1'b1;
						we_b   <= 1'b1;
						if(count_XOR_SBOX == 10'd1022) begin
							controlstate <= COST_1;
						end
						else begin													
							count_XOR_SBOX <= count_XOR_SBOX+10'd2;
							controlstate <= XOR_SBOX_1;
						end
					end
					else if(controlstate ==  COST_1)begin
						we_a   <= 1'b0;
						we_b   <= 1'b0;
						if(cost_r==32'd0)begin/*Quando custo chegar a zer0 e' porque acabou o loop*/
							/*Manda ler a primeira posicao*/
							index <= 10'd0;
							addr_a <= 10'd0;
							addr_b <= 10'd1;
							accessingMEM <= 1'b1;
							state <= PRE_OUT_HANDSHAKE;
						end 
						else begin
							pApB <= 64'h0000000000000000;
							count_XOR_PARRAY <= 5'd0;
							controlstate <=  XOR_PARRAY_1;
							state <= INIT_XOR;
							if(ctrl==1'b0)begin
								ctrl <= 1'b1;
							end else begin
								ctrl <= 1'b0;
								cost_r <= cost_r - 32'd1;
							end
						end
					end
				end
				else if(state == PREF) begin 
					L <= L^parray[count_round];//xL = xL ^ pArray[count_round];
					
					/*Ler duas posicoes da sboxes*/
					addr_a <= L[31:24]^parray[count_round][31:24];
					addr_b <= 10'd256+L[23:16]^parray[count_round][23:16];
					
					accessingMEM <= 1'b1;
					state <= F0;
				end
				else if (state == F0)begin
					if( (q_a+q_b) < 33'd4294967296)begin
						//F_r <= (sboxes[L[31:24]]+sboxes[10'd256+L[23:16]]) ^ sboxes[10'd512+L[15:8]];
						F_r <= (q_a+q_b);
					end 
					else begin
						F_r <= (q_a+q_b - 33'd4294967296);
					end
					/*Ler mais duas posicoes da sboxes*/
					addr_a <= 10'd512+L[15:8];
					addr_b <= 10'd768+L[7:0];
					accessingMEM <= 1'b1;
					state <= F2;
				end
				else if (state == F2)begin
					if( ( (F_r^q_a) + q_b) < 33'd4294967296)begin
						F_r <= ((F_r^q_a) + q_b);
						L <= R^((F_r^q_a) + q_b);
					end 
					else begin
						F_r <= ((F_r^q_a) + q_b) - 33'd4294967296;
						L <= R^(((F_r^q_a) + q_b) - 33'd4294967296);
					end
					R <= L;
					if(count_round<4'd15)begin
						count_round <= count_round+4'd1;
						state <=PREF;
					end
					else begin
						state <= XOR_P17_P18;
					end
				end
				else if(state == XOR_XR_SWAP)begin
					L <= R^F_r;
					R <= L;
					if(count_round<4'd15)begin
						count_round <= count_round+4'd1;
						state <=PREF;
					end
					else begin
						state <= XOR_P17_P18;
					end
				end
				else if(state == XOR_P17_P18) begin
					pApB[63:32] <=  R ^  parray[5'd17];
					pApB[31:0]  <=  L ^  parray[5'd16];
					state <= CONTROL_EXPAND_KEY;
				end
				else if(state == PRE_OUT_HANDSHAKE)begin
						VALID_S <= 1'b1;
						data_out_a <= q_a;
						data_out_b <= q_b;
						index_out  <=  index;
						index <= index + 10'd2;
						state <= OUT_HANDSHAKE;
				end
				else if(state == OUT_HANDSHAKE)begin
					if(READY_S==1'b1) begin
						VALID_S <= 1'b0;
						if(index_out != 10'd1022)begin
							addr_a <= index;
							addr_b <= index+10'd1;
							accessingMEM <= 1'b1;
							state <= PRE_OUT_HANDSHAKE;
						end 
						else begin
							READY_E <= 1'b1;
							state <= IN_HANDSHAKE;
						end
					end
				end	
			end 
			else begin
				/*leituras na memorira*/
				if(we_a) begin
					we_a <= 1'b0;
					we_b <= 1'b0;
				end
				if(state == IN_HANDSHAKE)begin
					READY_E <= 1'b1;
				end
				accessingMEM <= 1'b0;
			end
		end	
		else begin
			/* Reset everything here */
			count_round <= 4'd0;
			state    <=   IN_HANDSHAKE;
			READY_E   <=  1'b1;
			VALID_S   <=  1'b0;
			accessingMEM <= 1'b0;
		end
	end
endmodule
